#!/usr/bin/perl
#
# Limisc v1.0
#
# Software for Mini SSC-II device, for controlling servo motors
# via serial devices.
#
# This software is released with the GNU General Public License
# (read more at http://www.gnu.org/copyleft/gpl.html).
#
# (c) 2003 Rolf Johansson, <rojo@nocrew.org>.

use Curses;
use Curses::Widgets;
use Curses::Widgets::TextField;
use FileHandle;

init();
open_serial();

if($ARGV[0] eq "--run-program-once"){
    $loop=0;
    $runp=1;
}
if($ARGV[0] eq "--run-program-loop"){
    $loop=1;
    $runp=1;
}
if($runp){
    $interactive=0;
    load_program($ARGV[1]);
    if($pp > 0){
	print "Aligning servos to home positions...\n";
	for($i=0;$i<=$servos;$i++){
	    $servo[$i] = $servo_home[$i];
	}
	write_servopos();
	if($loop){
	    while(TRUE){
		for($row=0;$row<$pp;$row++){
		    print "Aligning servos to step $row...\n";
		    step_to_servopos($row);
		}
	    }
	}else{
	    for($row=0;$row<$pp;$row++){
		print "Aligning servos to step $row...\n";
		step_to_servopos($row);
	    }
	}
    }
    close_serial();
    exit;
}

$main = init_curses();

#$main->attrset(COLOR_PAIR(select_colour(qw(yellow black))));
#$main->attrset(0);
$maxl = LINES;
$maxl--;

#$main->box(ACS_VLINE, ACS_HLINE);

# Create/fill window for servo positions:

$servo_win = $main->subwin(10,COLS,0,0);
$servo_win->box(ACS_VLINE, ACS_HLINE);
$servo_win->standout();
$servo_win->addstr(0,1," Servo positions ");
$servo_win->standend();
$servo_win->addstr(1,1,"Max:");
$servo_win->addstr(3,1,"Min:");
for($i=0;$i<=$servos;$i++){
    $servo_win->addstr(1,($i*10)+5,"$servo_max[$i]");
    $servo_win->addstr(3,($i*10)+5,"$servo_min[$i]");
}
$servo_win->addstr(5,1,"LEFT/RIGHT: select servo, UP/DOWN: select position");
$servo_win->addstr(6,1,"PPAGE/NPAGE/HOME: select max/min/home positions");
$servo_win->addstr(8,1,"Commands: [A]dd [R]un [S]ave [L]oad [H]ome [Q]uit");

# Create the window for status and messages:

$status_win = $main->subwin(5,COLS,10,0);
$status_win->box(ACS_VLINE, ACS_HLINE);
$status_win->standout();
$status_win->addstr(0,1," Messages ");
$status_win->standend();

# Create the window for program:

$program_win = $main->subwin((LINES)-15,COLS,15,0);
$program_win->box(ACS_VLINE, ACS_HLINE);
$program_win->standout();
$program_win->addstr(0,1," Program ");
$program_win->standend();

status_add("Servos aligned to home positions.");
status_add("Waiting for user input.");

$main->refresh();

$smark=0;
$pp=0;

# Main loop:

while(!(uc($key) eq "Q")){
    &write_servopos();
    $servo_win->standout();
    $servo_win->addstr(2,($smark*10)+1,"S$smark: $servo[$smark]"
		       ." "x(6-length($servo[$smark])));
    $servo_win->standend();
    $servo_win->refresh();
    $key=scankey($main);
    if($key == KEY_HOME){
	$servo[$smark] = $servo_home[$smark];
    }
    if(uc($key) eq "H"){
	for($i=0;$i<=$servos;$i++){
	    $servo[$i] = $servo_home[$i];
	}
    }
    if(uc($key) eq "A"){
	$servopos[$pp]="$servo[0],$servo[1],$servo[2],$servo[3],$servo[4],";
	display_program();
	$pp++;
    }
    if(uc($key) eq "R"){
	if($pp > 0){
	    status_add("Running program:");
	    for($row=0;$row<$pp;$row++){
		status_add("Aligning servos to step $row...");
		step_to_servopos($row);
	    }
	}else{
	    status_add("No program to run.");
	}
    }
    if(uc($key) eq "S"){
	if($pp > 0){
	    $test_win = newwin(3,48,1,1);
	    $test_win->clear();
	    $test_win->box(ACS_VLINE, ACS_HLINE);
	    $test_win->standout();
	    $test_win->addstr(0,1," Save program... ");
	    $test_win->standend();
	    $test_win->addstr(1,1,"Filename: ");
	    $test_win->refresh();
	    echo();
	    curs_set(1);
	    $test_win->getstr($filename);
	    curs_set(0);
	    noecho();
	    $test_win->delwin();
	    $main->touchwin();
	    $main->redrawwin();
	    if($filename ne ""){
		if(open(FILE,">$ENV{'HOME'}/$filename")){
		    print FILE "# MiniSSC-II program - Servo Commander/Linux.\n";
		    for($row=0;$row<$pp;$row++){
			print FILE "$servopos[$row]\n";
		    }
		    close(FILE);
		    status_add("Program saved to \"$ENV{'HOME'}/$filename\".");
		}else{
		    status_add("Error: couldn't write to \"$ENV{'HOME'}/$filename\".");
		}
	    }else{
		status_add("Program save aborted.");
	    }
	}else{
	    status_add("No program to save.");
	}
    }
    if(uc($key) eq "L"){
	$test_win = newwin(3,48,1,1);
	$test_win->clear();
	$test_win->box(ACS_VLINE, ACS_HLINE);
	$test_win->standout();
	$test_win->addstr(0,1," Load program... ");
	$test_win->standend();
	$test_win->addstr(1,1,"Filename: ");
	$test_win->refresh();
	echo();
	curs_set(1);
	$test_win->getstr($filename);
	curs_set(0);
	noecho();
	$test_win->delwin();
	$main->touchwin();
	$main->redrawwin();
	load_program($filename);
    }
    if($key == KEY_NPAGE){
	$servo[$smark] = $servo_min[$smark];
    }
    if($key == KEY_PPAGE){
	$servo[$smark] = $servo_max[$smark];
    }
    if($key == KEY_LEFT){
	$smark--;
	if($smark == -1){
	    $smark = $servos;
	}
    }
    if($key == KEY_RIGHT){
	$smark++;
	if($smark > $servos){
	    $smark = 0;
	}
    }
    if($key == KEY_DOWN){
	$servo[$smark]--;
	if($servo[$smark] < $servo_min[$smark]){
	   $servo[$smark] = $servo_min[$smark];
	}
    }
    if($key == KEY_UP){
	$servo[$smark]++;
	if($servo[$smark] > $servo_max[$smark]){
	   $servo[$smark] = $servo_max[$smark];
	}
    }
    # Test area:
    if(uc($key) eq "!"){
	$test_win = newwin(15,40,5,5);
	$test_win->attrset(COLOR_PAIR(select_colour(qw(white blue))));
	$test_win->clear();
	$test_win->attrset(COLOR_PAIR(select_colour(qw(white blue))));
	$test_win->box(ACS_VLINE, ACS_HLINE);
	$test_win->standout();
	$test_win->addstr(0,1," Options ");
	$test_win->standend();
	$test_win->attrset(COLOR_PAIR(select_colour(qw(white blue))));
	$test_win->addstr(2,1,"Serial device:");
	$test_win->addstr(3,1,"Number of servos:");
	$test_win->refresh();
	$test_win->attrset(0);
	$test_win_key=scankey($test_win);
	$test_win->delwin();
	#$servo_win->refresh();
	#$status_win->clear();
	#$status_win->refresh();
	#$program_win->refresh();
	#$servo_win->touchwin();
	#$servo_win->redrawwin();
	#$servo_win->refresh();
	#$status_win->redrawwin();
	$main->refresh();
	$main->redrawwin();
    }
}

endwin();
close_serial();
exit;

sub status_add{
    my($text)=@_;
    $statustext[$smsgno] = "$text";
    if($smsgno > 2){
	$status_win->addstr(1,1,$statustext[$smsgno-2]
			    ." "x((COLS)-length($statustext[$smsgno-2])-2));
	$status_win->addstr(2,1,$statustext[$smsgno-1]
			    ." "x((COLS)-length($statustext[$smsgno-1])-2));
	$status_win->addstr(3,1,$statustext[$smsgno]
			    ." "x((COLS)-length($statustext[$smsgno])-2));
    }else{
	$status_win->addstr($smsgno+1,1,$statustext[$smsgno]);
    }
    $smsgno++;
    $status_win->refresh();
}
sub write_servopos{
    for($i=0;$i<=$servos;$i++){
	if($interactive){
	    $servo_win->addstr(2,($i*10)+1,"S$i: $servo[$i]"
			       ." "x(6-length($servo[$i])));
	}
	printf(SERIAL "%c%c%c",255,$i,$servo[$i]);
    }
    if($interactive){
	$servo_win->refresh();
    }
}
sub step_to_servopos{
    my($pline)=@_;
    my($servos_moved,$currservo)=0;
    my($i,$servo_done,$servo_to,$servo_now);
    # First, grab the current positions:
    for($i=0;$i<=$servos;$i++){
	$servo_now[$i]=$servo[$i];
	$servo_done[$i]=0;
    }
    # Second, step to the positions given in program row "$pline":
    ($servo_to[0],$servo_to[1],$servo_to[2],$servo_to[3],$servo_to[4])
	= split(/\,/,$servopos[$pline]);
    # Third, step to the positions (one step per servo at a time):
    if(!$interactive){
	print "-> $servopos[$pline]\n";
    }
    while(!$servos_moved){
	if($servo_to[$currservo] == $servo_now[$currservo] &&
	   $servo_done[$currservo] == 0){
	    $servo_done[$currservo] = 1;
	}
	if($servo_to[$currservo] > $servo_now[$currservo] &&
	   $servo_done[$currservo] == 0){
	    $servo[$currservo]++;
	    write_servopos();
	}
	if($servo_to[$currservo] < $servo_now[$currservo] &&
	   $servo_done[$currservo] == 0){
	    $servo[$currservo]--;
	    write_servopos();
	}
	if($servo_to[$currservo] == $servo[$currservo]){
	    $servo_done[$currservo] = 1;
	}
	if($servo_done[0] && $servo_done[1] && $servo_done[2]
	   && $servo_done[3] && $servo_done[4]){
	    $servos_moved = 1;
	}
	for($b=0;$b<150;$b++){
	    for($c=0;$c<5;$c++){
		$d++;
	    }
	}
	if($currservo == $servos){
	    $currservo = 0;
	}else{
	    $currservo++;
	}
    }
}
sub load_program{
    my($filename)=@_;
    if($filename ne ""){
	if(open(FILE,"$filename")){
	    $pp=0;
	    if($interactive){
		$program_win->refresh();
	    }
	    while($row = <FILE>){
		chop($row);
		if($row !~ /^#/ && $row ne ""){
		   $servopos[$pp]=$row;
		   if($interactive){
		       display_program();
		   }
		   $pp++;
	       }
	    }
	    close(FILE);
	    if($interactive){
		status_add("Program loaded from \"$filename\".");
	    }else{
		print "Program loaded from \"$filename\".\n";
	    }
	}else{
	    if($interactive){
		status_add("Error: couldn't read \"$filename\".");
	    }else{
		print "Error: couldn't read \"$filename\".\n";
	    }
	}
    }else{
	if($interactive){
	    status_add("Program load aborted.");
	}else{
	    print "Error: No filename supplied, program load aborted.\n";
	}
    }
}
sub display_program{
    my($i,$s,$n);
    $program_win->addstr(1,1,"Step");
    $program_win->addstr(1,10,"Servo 0");
    $program_win->addstr(1,20,"Servo 1");
    $program_win->addstr(1,30,"Servo 2");
    $program_win->addstr(1,40,"Servo 3");
    $program_win->addstr(1,50,"Servo 4");
    $program_win->addstr(1,60,"");
    $program_win->refresh();
    for($i=0;$i<=$pp;$i++){
	($s[0],$s[1],$s[2],$s[3],$s[4],$sc)=split(/\,/,$servopos[$i]);
	$program_win->addstr(2+$i,1," "x((COLS)-2));
	$program_win->addstr(2+$i,1,"$i");
	for($n=0;$n<=$servos;$n++){
	    $program_win->addstr(2+$i,($n*10)+10,"$s[$n]");
	}
	$program_win->addstr(2+$i,60,"$sc");
    }
    $program_win->refresh();
}
sub open_serial{
    open(SERIAL,"+< $serial_device") or die "open: $!";
    SERIAL->autoflush(1);
}
sub close_serial{
    close(SERIAL);
}
sub init{
    $serial_device = "/dev/ttyS0";
    $servos = 4;
    # These min/max settings are for the 180 degree range
    # of motion (R jumper set to "on", on the MiniSSC-II):
    $servo_min[0]  = 30;
    $servo_max[0]  = 254;
    $servo_home[0] = 138;
    $servo_min[1]  = 40;
    $servo_max[1]  = 180;
    $servo_home[1] = 178;
    $servo_min[2]  = 30;
    $servo_max[2]  = 205;
    $servo_home[2] = 205;
    $servo_min[3]  = 30;
    $servo_max[3]  = 230;
    $servo_home[3] = 167;
    $servo_min[4]  = 0;
    $servo_max[4]  = 100;
    $servo_home[4] = 100;
    # Set the default values to the home positions:
    for($i=0;$i<=$servos;$i++){
	$servo[$i] = $servo_home[$i];
    }
    $smsgno = 0;
    $interactive=1;
}
sub init_curses{
    my($tmp) = new Curses;
    noecho();
    halfdelay(5);
    $tmp->keypad(1);
    curs_set(0);
    return $tmp;
}
